from . import config
import logging
import multiprocessing
import base64
import os
import requests
import time
from .progressbar import ProgressBar
import yaml
from .data import Model
from .config import *
import traceback
from pprint import pformat

def download_model(arg_tuple):
    model_id, model_run_info = arg_tuple
    try:
        model_json = requests.get(MDB_MODEL_DOWNLOAD_URL.format(model_id=model_id)).json()
        model = Model(
            *(
                model_json[key]
                for key in ("object_id", "object_name", "object_created", "object_ver_date")
            )
        )
        url = None
        for att in model_json["object_attribute_values"]:
            if att["attribute_id"] == 23:
                url = att["value"]
                break
        # print(model.id)
        model_zip_uri = os.path.join(
            MODELS_ZIP_DIR, "{model_id}.zip".format(model_id=model.id)
        )
        with open(model_zip_uri, "wb+") as zipfile:
            zipfile.write(base64.standard_b64decode(url["file_content"]))

        if "github" in model_run_info:
            # This means we should try to replace the version of the model that
            # we downloaded from the ModelDB API just above with a version from
            # GitHub
            github = model_run_info["github"]
            organisation = "ModelDBRepository"
            suffix = "" # default branch
            if github == "default":
                pass
            elif github.startswith("pull/"):
                pr_number = int(github[5:])
                suffix = "/pull/{}/head".format(pr_number)
            elif github.startswith('/'):
                # /org implies that we use the default branch from org/model_id
                organisation = github[1:]
            else:
                raise Exception("Invalid value for github key: {}".format(github))
            github_url = "https://api.github.com/repos/{organisation}/{model_id}/zipball{suffix}".format(
                model_id=model_id, organisation=organisation, suffix=suffix
            )
            # Replace the local file `model_zip_uri` with the zip file we
            # downloaded from `github_url`
            num_attempts = 3
            status_codes = []
            for _ in range(num_attempts):
                github_response = requests.get(github_url)
                status_codes.append(github_response.status_code)
                if github_response.status_code == requests.codes.ok:
                    break
                time.sleep(5)
            else:
                raise Exception(
                    "Failed to download {} with status codes {}".format(
                        github_url, status_codes
                    )
                )
            with open(model_zip_uri, "wb+") as zipfile:
                zipfile.write(github_response.content)
    except Exception as e:  #  noqa
        model = e

    return model_id, model


class ModelDB(object):
    metadata = property(lambda self: self._metadata)
    run_instr = property(lambda self: self._run_instr)

    def __init__(self):
        self._metadata = {}
        self._run_instr = {}

        self._load_run_instructions()
        try:
            self._load_metadata()
        except FileNotFoundError:
            logging.warning(
                "{} not found!".format(MODELDB_METADATA_FILE)
            )
        except yaml.YAMLError as y:
            logging.error("Error loading {}: {}".format(MODELDB_METADATA_FILE, y))
            raise y
        except Exception as e:
            raise e

    def _download_models(self, model_list=None):
        if not os.path.isdir(MODELS_ZIP_DIR):
            logging.info("Creating cache directory: {}".format(MODELS_ZIP_DIR))
            os.mkdir(MODELS_ZIP_DIR)
        models = requests.get(MDB_NEURON_MODELS_URL).json() if model_list is None else model_list
        pool = multiprocessing.Pool()
        processed_models = pool.imap_unordered(
            download_model,
            [(model_id, self._run_instr.get(model_id, {})) for model_id in models],
        )
        download_err = {}
        for model_id, model in ProgressBar.iter(processed_models, len(models)):
            if not isinstance(model, Exception):
                self._metadata[model_id] = model
            else:
                download_err[model_id] = model

        if download_err:
            logging.error("Error downloading models:")
            logging.error(pformat(download_err))

        self._save_metadata()

    def _load_metadata(self):
        with open(MODELDB_METADATA_FILE) as meta_file:
            self._metadata = yaml.load(meta_file, yaml.Loader)

    def _load_run_instructions(self):
        with open(MODELDB_RUN_FILE) as run_file:
            self._run_instr = yaml.load(run_file, yaml.Loader)

    def _save_metadata(self):
        with open(MODELDB_METADATA_FILE, "w+") as meta_file:
            yaml.dump(self._metadata, meta_file, sort_keys=True)

    def download_models(self, model_list=None):
        if model_list is None:
            try:
                os.remove(MODELDB_METADATA_FILE)
            except OSError:
                pass
        self._download_models(model_list)

    # TODO -> check/update models
    def update_models(self):
        pass




driver_hoc_header = r"""
strdef verify_dir_ 
verify_dir_ = "{model_dir}"
"""

driver_hoc_body = r"""
if (name_declared("verify_dir_") == 0) {
	execute("~strdef verify_dir_")
	execute("verify_dir_ = \".\"")
}
strdef verify_tstr_
objref verify_glist_, verify_xvec_, verify_yvec_, verify_file_
verify_file_ = new File()
sprint(verify_tstr_, "%s/gout", verify_dir_)
verify_file_.wopen(verify_tstr_)
verify_xvec_ = new Vector()
verify_yvec_ = new Vector()
verify_glist_ = new List("Graph")

proc verify_graph_() {local i, j, k
	verify_file_.printf("Graphs %d\n", verify_glist_.count)
	for i=0, verify_glist_.count-1 {
		verify_file_.printf("%s\n", verify_glist_.object(i))
		k = 0
		for (j=-1; (j=verify_glist_.object(i).line_info(j, verify_xvec_)) != -1; ){
			k += 1
		}
		verify_file_.printf("lines %d\n", k)
		for (j=-1; (j=verify_glist_.object(i).getline(j, verify_xvec_, verify_yvec_)) != -1; ){
			verify_file_.printf("points %d\n", verify_xvec_.size)
			verify_file_.printf("xvec%d\n", j)
			verify_xvec_.printf(verify_file_)
			verify_file_.printf("yvec%d\n", j)
			verify_yvec_.printf(verify_file_)
		}
	}
}
"""

quit_hoc = r"""
quit()
"""



import logging
import multiprocessing
import subprocess
import platform
import sys
from .progressbar import ProgressBar
from .data import Model
from . import modeldb
from .config import *
from .hocscripts import *
import zipfile
import glob
import traceback
import json
import time
import shutil
import yaml

ModelDB = modeldb.ModelDB()


def is_dir_non_empty(directory):
    """
    Returns True if the `directory` exists and is non-empty
    """
    try:
        if any(os.scandir(directory)):
            return True
    except Exception:  # noqa
        pass

    return False


class ModelRun(dict):
    def __init__(self, model, working_dir, clean=False, norun=False, inplace=False):
        super().__init__()
        self._model = model
        self._working_dir = os.path.abspath(working_dir)
        self._logs = []
        self._gout = []
        self._nrn_run = []
        self._nrn_run_error = False
        self._no_mosinit_hoc = False
        self._run_time = 0
        self._run_times = {}
        self._run_py = False
        self._clean = clean
        self._norun = norun
        self._inplace = inplace

        self["run_info"] = {}

        self._fetch_model()

    def _fetch_model(self):
        # get run instruction from ModelDB
        if self.id in ModelDB.run_instr:
            self.update(ModelDB.run_instr[self.id])

        # check if model is run in Python
        if "python" in self:
            self._run_py = True
        # if no actual run instructions are specified, add default run command
        if "run" not in self:
            if self.run_py:
                self["run"] = ["python mosinit.py"]
            else:
                self["run"] = ["verify_graph_()"]

        if self._norun:
            self["norun"] = True

        if self._inplace:
            self["inplace"] = True

    run_info = property(lambda self: self["run_info"])

    logs = property(lambda self: self._logs)
    gout = property(lambda self: self._gout)
    nrn_run = property(lambda self: self._nrn_run)
    no_mosinit_hoc = property(lambda self: self._no_mosinit_hoc)
    run_py = property(lambda self: self._run_py)

    nrn_run_error = property(lambda self: self._nrn_run_error)
    model_dir = property(lambda self: self.run_info["model_dir"] if "model_dir" in self.run_info else "")
    working_dir = property(lambda self: self._working_dir)
    run_time = property(lambda self: self._run_time)
    run_times = property(lambda self: self._run_times)

    id = property(lambda self: self._model.id)


def curate_log_string(model, logstr):
    return logstr.replace(model.model_dir, '%model_dir%') if len(model.model_dir) else logstr


def append_log(model, model_sink, text):
    model_sink.extend(curate_log_string(model, text).split('\n'))


def run_commands(model, cmds, env={}, work_dir=None):
    full_env = os.environ
    full_env.update(env)
    out, _ = subprocess.Popen(
            cmds,
            env=full_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            cwd=model.model_dir if work_dir is None else work_dir,
        ).communicate()

    model.logs.extend(curate_log_string(model, out).split('\n'))


def run_neuron_cmds(model, cmds):
    sp = subprocess.Popen(
        cmds,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=model.run_info["start_dir"],
    )
    out, _ = sp.communicate()
    try:
        out = out.decode("utf-8")
    except UnicodeDecodeError:
        raise Exception("Could not decode output:" + repr(out))
    model.nrn_run.extend(curate_log_string(model, out).splitlines())
    if sp.returncode > 1:
        model._nrn_run_error = True


def clean_model_dir(model):
    # delete x86_64 folder
    run_commands(model, ["/bin/sh", "-c", "rm -rf ./{}/".format(platform.machine())], work_dir=model.run_info["start_dir"])


def compile_mods(model, mods):
    # Unfortunately nrnivmodl doesn't have an option to steer how much build
    # parallellism it tries to do, it just hardcodes `make -j 4`. Because we
    # parallelise over models, at a higher level, we want to remove this
    # internal parallelism from nrnivmodl. In the CI we install NEURON using
    # pip from precompiled wheels, and the real nrnivmodl is hidden behind
    # an extra layer of wrappers. This makes it inconvenient to change the
    # hardcoded value. Instead, we try to achieve the same effect using Make's
    # environment variables. --max-load 0.0 should ban >1 job being launched if
    # the system load is larger than zero.
    run_commands(
        model,
        ["nrnivmodl"] + mods,
        env={"MAKEFLAGS": " --max-load 0.0"},
        work_dir=model.run_info["start_dir"],
    )


def build_driver_hoc(model):
    model.run_info["driver"] = os.path.join(model.model_dir, "driver.hoc")

    with open(model.run_info["driver"], "w") as drv:
        drv.write(driver_hoc_header.format(model_dir=model.model_dir))
        drv.writelines(driver_hoc_body)
        if model["run"] is not None:
            drv.writelines("\n".join(model["run"]))  # how to run the model


def build_quit_hoc(model):
    model.run_info["start_dir"] = model.model_dir

    with open(os.path.join(model.model_dir, "quit.hoc"), "w") as qt:
        qt.write(quit_hoc)

    model.run_info["init"] = os.path.join(model.model_dir, "quit.hoc")


def select_mosinit(model):
    # look for `mosinit.hoc`. It could also be produced by `init` script above
    mosfiles = glob.glob(model.model_dir + "/**/mosinit.hoc", recursive=True)
    mosfiles.sort()
    if len(mosfiles):
        model.run_info["start_dir"] = os.path.dirname(os.path.join(model.model_dir, mosfiles[0]))
        model.run_info["init"] = mosfiles[0]
    else:
        build_quit_hoc(model)
        model._no_mosinit_hoc = True


def build_and_run_script(model):
    if "script" in model:
        with open(os.path.join(model.model_dir, "script.tmp"), "w") as script:
            script.writelines("\n".join(model["script"]))
            script.flush()
        run_commands(model, ["/bin/sh", "script.tmp"])
        model.run_info["script"] = model["script"]


def build_python_runfile(model):
    model.run_info["start_dir"] = model.model_dir
    with open(os.path.join(model.model_dir, "model_run.py"), "w") as script:
        script.writelines("\n".join(model["run"]))
        script.flush()
    model.run_info["init"] = "model_run.py"
    model.run_info["driver"] = ""


def prepare_model(model):
    # unzip model from cache
    with zipfile.ZipFile(
            os.path.join(MODELS_ZIP_DIR, str(model.id) + ".zip"), "r"
    ) as zip_ref:
        model_dir = os.path.join(
            model.working_dir,
            str(model.id),
            os.path.dirname(zip_ref.infolist()[0].filename),
        )
        model_run_info_file = os.path.join(model_dir, str(model.id) + ".yaml")
        if model._inplace and os.path.isfile(model_run_info_file):
            with open(model_run_info_file) as run_info_file:
                model["run_info"] = yaml.load(run_info_file, yaml.Loader)
        else:
            if model._clean and is_dir_non_empty(model_dir):
                shutil.rmtree(model_dir)
            zip_ref.extractall(os.path.join(model.working_dir, str(model.id)))

            # set model_dir
            model.run_info["model_dir"] = model_dir

            # write driver.hoc
            build_driver_hoc(model)

            # write and execute extra script if specified in the run instructions
            build_and_run_script(model)

            # Determine init file: HOC or Python.
            # Python
            if model.run_py:
                build_python_runfile(model)
            # HOC: If 'run' is None -> quit.hoc (DoNotRun = yes) else look for 'mosinit.hoc'
            elif model["run"] is None:
                build_quit_hoc(model)
            else:
                select_mosinit(model)

            # dump run_info into model_dir
            with open(model_run_info_file, "w+") as run_info_file:
                yaml.dump(model.run_info, run_info_file, sort_keys=True)


def run_model(model):
    start_time = time.perf_counter()
    # Some models are skipped on purpose
    if "skip" in model:
        append_log(model, model.logs,
            "Model is skipped according to modeldb-run.yaml:\n\t{}\n".format(
                model["comment"]
            )
        )
        return model
    mods = None

    try:
        # prepare model
        prepare_model(model)

        if model["run"] is None:
            append_log(model, model.logs,
                       "Model in do not run mode according to modeldb-run.yaml:\n\t{}\n".format(
                           model["comment"]
                       )
           )

        # Get mod files. Model can have a custom mod directory or directories, otherwise we search in the start_dir
        if "model_dir" in model:
            mods = []
            for moddir in model["model_dir"].split(';'):
                mod_pattern = os.path.join(model.model_dir, moddir) + "/*.mod"
                mods.extend(glob.glob(mod_pattern))
        else:
            mod_pattern = model.run_info["start_dir"] + "/*.mod"
            mods = glob.glob(mod_pattern)
        # compile mods if available
        if len(mods):
            # in case of reruns
            clean_model_dir(model)
            # translate them to cpp
            compile_mods(model, mods)

    except Exception:  # noqa
        append_log(model, model.logs, traceback.format_exc())

    # Record how long the preparation took (even if it failed)
    stop_time = time.perf_counter()
    model._run_times["nrnivmodl"] = stop_time - start_time
    start_time = stop_time

    # run NEURON
    if "norun" in model:
        append_log(model, model.logs,
            "Model is not run due to --norun option"
        )
    else:
        try:
            nrn_exe = "./{}/special".format(platform.machine()) if mods is not None and len(mods) else "nrniv"
            # '-nogui' creates segfault
            model_run_cmds = [nrn_exe, '-nobanner']
            if model.run_py:
                model_run_cmds.append('-python')
            model_run_cmds += [model.run_info["init"], model.run_info["driver"]]
            append_log(model, model.nrn_run, "RUNNING -> {}".format(" ".join(model_run_cmds)))
            run_neuron_cmds(model, model_run_cmds)
            if os.path.isfile(os.path.join(model.model_dir, "gout")):
                with open(os.path.join(model.model_dir, "gout"), 'r') as gout:
                    model._gout = gout.readlines()
        except Exception:  # noqa
            append_log(model, model.nrn_run, traceback.format_exc())
            model._nrn_run_error = True

    stop_time = time.perf_counter()
    model._run_times["model"] = stop_time - start_time

    # Record the total too (for backwards compatibility)
    model._run_time = str(sum(model._run_times.values()))

    return model


class ModelRunManager(object):
    def __init__(self, master_dir, gout=False, clean=False, norun=False, inplace=False):
        self.master_dir = master_dir
        self.logfile = str(master_dir) + ".log"
        self.dumpfile = str(master_dir) + ".json"
        self._setup_logging()
        self.logger.info(
            "Initialized -> logfile: " + self.logfile
        )
        self.run_logs = {}
        self._gout = gout
        self._clean = clean
        self._norun = norun
        self._inplace = inplace

    def _setup_logging(self):
        self.logger = logging.getLogger("dev")
        self.logger.setLevel(logging.DEBUG)
        self.logFormatter = logging.Formatter(
            fmt="%(asctime)s :: %(levelname)-8s :: %(message)s"
        )

        self.fileHandler = logging.FileHandler(self.logfile)
        self.fileHandler.setFormatter(self.logFormatter)
        self.fileHandler.setLevel(logging.DEBUG)

        self.consoleHandler = logging.StreamHandler()
        self.consoleHandler.setFormatter(self.logFormatter)
        self.consoleHandler.setLevel(logging.INFO)

        self.logger.addHandler(self.fileHandler)
        self.logger.addHandler(self.consoleHandler)

    def _grep_for_errors(self):
        if len(self.run_logs):
            self.logger.info("Grepping all models for ' error:' and dumping run errors")

            for model_id, logs in self.run_logs.items():
                mod_errors = list(filter(lambda x: " error:" in x, logs["logs"]))
                if len(mod_errors) > 0:
                    self.run_logs[model_id]["moderr"] = mod_errors
                    self.logger.error(
                        str(model_id) + "\n\t" + "\t".join(mod_errors)
                    )
                if "nrn_run_err" in logs and logs["nrn_run_err"] is True:
                    self.logger.error(
                        str(model_id) + "\n\t" + "\t".join(logs["nrn_run"])
                    )

    def _dump_run(self):
        self.logger.info("Dumping run logs to {} ...".format(self.dumpfile))

        # Run info, use key 0
        self.run_logs[0] = {}
        from neuron import __version__ as nrn_ver
        self.run_logs[0]["NEURON version"] = nrn_ver
        self.run_logs[0]["Stats"] = self._run_stats(self.run_logs)

        # Dump logs
        with open(self.dumpfile, "w+") as dump_file:
            json.dump(self.run_logs, dump_file, indent=4, sort_keys=True)

    @staticmethod
    def _run_stats(json_report):
        stats = {}
        stats["Total nof models run"] = len(json_report) - 1  # discard 0
        failed_mods = []
        failed_runs = []
        skipped_runs = []
        
        for model_id in json_report.keys():
            # look for models that are marked `skip` in `modeldb-run.yaml`
            if "do_not_run" in json_report[model_id]:
                skipped_runs.append(model_id)

            # moderr happens if mods present and nrnivmodl failed
            # if no moderr we look for nrn_run_err    
            if "moderr" in json_report[model_id]:
                failed_mods.append(model_id)
            elif "nrn_run_err" in json_report[model_id]:
                failed_runs.append(model_id)

        stats["Failed models"] = {"Count": len(failed_mods),
                                "Accession numbers": failed_mods}

        stats["Failed runs"] = {"Count": len(failed_runs),
                                "Accession numbers": failed_runs}

        stats["Skipped runs"] = {"Count": len(skipped_runs),
                                "Accession numbers": skipped_runs}

        return stats


    def _run_models(self, model_runs):
        pool = multiprocessing.Pool()

        processed_models = pool.imap_unordered(run_model, model_runs)
        for model in ProgressBar.iter(processed_models, self.nof_models):
            self.run_logs[model.id] = {}
            self.run_logs[model.id]["logs"] = model.logs
            if self._gout:
                self.run_logs[model.id]["gout"] = model.gout
            self.run_logs[model.id]["nrn_run"] = model.nrn_run
            if "skip" in model:
                self.run_logs[model.id]["do_not_run"] = True
            if model.nrn_run_error:
                self.run_logs[model.id]["nrn_run_err"] = True
            if model.no_mosinit_hoc:
                self.run_logs[model.id]["no_mosinit_hoc"] = True
            self.run_logs[model.id]["run_info"] = model.run_info
            self.run_logs[model.id]["run_time"] = model.run_time
            self.run_logs[model.id]["run_times"] = model.run_times
            self.logger.debug("Done for: {} in {}".format(str(model.id), str(model.run_times)))

        self._grep_for_errors()
        self._dump_run()

    def run_models(self, model_list=None):
        self.logger.info(
            "Master directory is: " + self.master_dir
        )

        if not os.path.isdir(self.master_dir):
            self.logger.info("Creating master directory...")
            os.mkdir(self.master_dir)

        # models selection
        models_selected = (
            ModelDB.metadata.values()
            if model_list is None
            else (ModelDB.metadata[k] for k in model_list)
        )

        # prepare ModelRun objects
        models_to_run = (
            ModelRun(mdl, self.master_dir, self._clean, self._norun, self._inplace) for mdl in models_selected
        )

        # number of models
        self.nof_models = (
            len(ModelDB.metadata) if model_list is None else len(model_list)
        )

        self.logger.info("Running models ...")
        self.logger.info(
            "\t\t-> number of models: " + str(self.nof_models)
        )
        self._run_models(models_to_run)
        self.logger.info("Done.")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(
            "Need to specify master directory " "containing the different model folders"
        )
        sys.exit(1)

    ModelRunManager(sys.argv[1]).run_models()



import json
import re
import difflib
import logging
import os
import shlex
import subprocess
from .modeldb import ModelDB
from pygments import highlight
from pygments.lexers import DiffLexer
from pygments.formatters import HtmlFormatter


mdb = ModelDB()


def curate_run_data(run_data, model=None):
    curated_data = run_data
    
    regex_dict = {
        # /../nrniv: Assignment to modern physical constant FARADAY	<-> ./x86_64/special: Assignment to modern physical constant FARADAY
        "^/.*?/nrniv:": "%neuron-executable%:",
        "^\\./x86_64/special:": "%neuron-executable%:",
        # nrniv: unable to open font "*helvetica-medium-r-normal*--14*", using "fixed" <-> special: unableto open font "*helvetica-medium-r-normal*--14*", using "fixed"
        "^nrniv:": "%neuron-executable%:",
        "^special:": "%neuron-executable%:",
        "(Mon|Tue|Wed|Thu|Fri|Sat|Sun) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+\d+\s+\d+:\d+:\d+ [A-Z\s]+ \d+": "%date_command%",
        "total run time [0-9\.]+": "total run time %run_time%",
        "(^.*distutils.*$)": "",
        "/.*?/lib/python.*/site-packages/" : "%python-site-packages%",
    }
    
    for model_specific_substitution in mdb.run_instr.get(model, {}).get("curate_patterns", []):
        regex_dict[model_specific_substitution["pattern"]] = model_specific_substitution["repl"]

    for regex_key, regex_value in regex_dict.items():
        updated_data = []
        pattern = re.compile(regex_key)
        for line in curated_data:
            new_line, number_of_subs = pattern.subn(regex_value, line)
            if number_of_subs:
                logging.debug("{} matched {} time(s)".format(regex_key, number_of_subs))
                logging.debug("{} -> {}".format(line, new_line))
            # if we are replacing a full line with an empty string, don't add it to the curated data
            if new_line:
                updated_data.append(new_line)
        curated_data = updated_data

    return curated_data


def diff_reports(report1_json, report2_json):
    diff_dict = {}
    gout_dict = {}
    runtime_dict = {}

    with open(report1_json, 'r+') as f, open(report2_json, 'r+') as f2:
        data_a = json.load(f)
        data_b = json.load(f2)

        hd = difflib.HtmlDiff()
        v1 = data_a["0"]["NEURON version"]
        v2 = data_b["0"]["NEURON version"]
        diff_dict["0"] = hd.make_table(
            json.dumps(data_a["0"], indent="\t").split("\n"),
            json.dumps(data_b["0"], indent="\t").split("\n"),
        ).replace("\n", "")
        stats_dict = {v1: data_a["0"]["Stats"], v2: data_b["0"]["Stats"]}
        for k in data_a.keys():
            if int(k) == 0:
                continue  # skip info key
            if k not in data_b:
                ud_empty = difflib.unified_diff(data_a[k]["nrn_run"], ["Accession number {} not found in report2".format(k)])
                diff_dict[k] = highlight('\n'.join(ud_empty), DiffLexer(), HtmlFormatter(linenos=True, cssclass="colorful", full=True))
                continue            
            curated_a = curate_run_data(data_a[k]["nrn_run"], model=int(k))
            curated_b = curate_run_data(data_b[k]["nrn_run"], model=int(k))
            start_dir_a = data_a[k]["run_info"]["start_dir"] if "run_info" in data_a[k] and "start_dir" in data_a[k]["run_info"] else "unknown"
            start_dir_b = data_b[k]["run_info"]["start_dir"] if "run_info" in data_b[k] and "start_dir" in data_b[k]["run_info"] else "unknown"
            if curated_a != curated_b:
                ud = difflib.unified_diff(curated_a, curated_b,  fromfile=start_dir_a,
                                             tofile=start_dir_b)
                diff_dict[k] = highlight('\n'.join(ud), DiffLexer(), HtmlFormatter(linenos=True, cssclass="colorful", full=True))
                
            def _speedup(a, b):
                dict = {}
                dict["v1"] = a
                dict["v2"] = b
                # compute slowdown/speedup relative to runtime_b (negative means slowdown)
                dict["speedup"] = (float(b) - float(a)) / float(b) * 100
                return dict

            # List of keys that make gout comparison and speedup comparison pointless
            skip_keys = {"do_not_run", "moderr", "nrn_run_err"}
            if skip_keys.isdisjoint(data_a[k]) and skip_keys.isdisjoint(data_b[k]):
                # compare runtimes and compute slowdown or speedup
                runtime_dict[k] = {}
                runtime_dict[k]["total"] = _speedup(data_a[k]["run_time"], data_b[k]["run_time"])
                for runkey in ("model", "nrnivmodl"):
                    if runkey in data_a[k]["run_times"] and runkey in data_b[k]["run_times"]:
                        runtime_dict[k][runkey] = _speedup(data_a[k]["run_times"][runkey], data_b[k]["run_times"][runkey])
                
                # compare gout
                gout_a_file = os.path.join(data_a[k]["run_info"]["start_dir"], "gout")
                gout_b_file = os.path.join(data_b[k]["run_info"]["start_dir"], "gout")
                # gout may be missing in one of the paths. `diff -N` treats non-existent files as empty.
                if os.path.isfile(gout_a_file) or os.path.isfile(gout_b_file):
                    diff_out = subprocess.getoutput(
                        "diff -uN --speed-large-files {} {} | head -n 30".format(
                            shlex.quote(gout_a_file), shlex.quote(gout_b_file)
                        )
                    )
                    if diff_out:
                        gout_dict[k] = highlight(diff_out, DiffLexer(), HtmlFormatter(linenos=True, cssclass="colorful", full=True))

    return diff_dict, gout_dict, runtime_dict, stats_dict, v1, v2



"""Global configuration variables used in the project"""

import os

# MDB_NEURON_MODELS_URL = "https://senselab.med.yale.edu/_site/webapi/object.json/?cl=19&oid=1882"
MDB_NEURON_MODELS_URL = (
    "http://modeldb.science/api/v1/models?modeling_application=NEURON"
)
MDB_MODEL_DOWNLOAD_URL = (
    "https://senselab.med.yale.edu/_site/webapi/object.json/{model_id}"
)

ROOT_DIR = os.path.abspath(__file__ + "/../../")

MODELS_ZIP_DIR = "%s/cache" % ROOT_DIR
MODELDB_ROOT_DIR = "%s/modeldb" % ROOT_DIR
MODELDB_METADATA_FILE = "%s/modeldb-meta.yaml" % MODELDB_ROOT_DIR
MODELDB_RUN_FILE = "%s/modeldb-run.yaml" % MODELDB_ROOT_DIR


class Model(object):
    def __init__(self, object_id, name, created, ver_date):
        self._object_id = object_id
        self._name = name
        self._created = created
        self._ver_date = ver_date

    @property
    def id(self):
        return self._object_id

    name = property(lambda self: self._name)
    created = property(lambda self: self._created)
    last_modified = property(lambda self: self._ver_date)



import logging
from docopt import docopt
from os.path import abspath
from pprint import pprint
from . import config
from .modelrun import ModelRunManager, is_dir_non_empty
from .modeldb import ModelDB
import inspect
import shlex
import subprocess
from jinja2 import Environment, FileSystemLoader
import os
import sys
from .config import *
from pathlib import Path
from .report import diff_reports
import json
import shutil


def runmodels(args=None):
    """runmodels

    Run nrn-modeldb-ci for all or specified models

    Usage:
        runmodels --workdir=<PATH> [options] [<model_id>...]
        runmodels -h    Print help

    Arguments:
        --workdir=<PATH>          Required: directory where to run the models and store the reports
        model_id=<n>            Optional: ModelDB accession number(s) to run; default is all available models

    Options:
        --gout                  Include gout into the report. Note that gout data can be very big, so disabled by default.
        --virtual               Run in headless mode. You need a back-end like Xvfb.
        --clean                 Auto-clean model working directory before running (useful for consecutive runs and failsafe)
        --norun                 Compile and link only (nrnivmodl).
        --inplace               Skip model preparation logic, simply run NEURON.

    Examples
        runmodels --workdir=/path/to/workdir                        # run all models
        runmodels --clean --workdir=/path/to/workdir 23613 12344    # run models 23613 & 12344
    """
    options = docopt(runmodels.__doc__, args)
    working_dir = options.pop("--workdir")
    model_ids = [int(model_id) for model_id in options.pop("<model_id>")]
    gout = options.pop("--gout", False)
    virtual = options.pop("--virtual", False)
    clean = options.pop("--clean", False)
    norun = options.pop("--norun", False)
    inplace = options.pop("--inplace", False)

    if os.path.abspath(working_dir) == ROOT_DIR:
        print("Cannot run models directly into nrn-modeldb-ci ROOT_DIR -> {}".format(ROOT_DIR))
        sys.exit(1)

    if clean and inplace:
        print("ERROR: --clean and --inplace are mutually exclusive")
        sys.exit(1)

    if not (clean or inplace) and is_dir_non_empty(working_dir):
        print("ERROR: WorkingDirectory {} exists and is non empty.".format(working_dir))
        print("\t re-run with one of these options:\n"
              "\t\t--clean \t-> if you wish to OVERWRITE model runs (delete content from --workdir and re-build from cache)\n"
              "\t\t--inplace \t-> if you wish to re-run the same model (content in --workdir is kept)\n")
        sys.exit(1)

    mrm = ModelRunManager(working_dir, gout=gout, clean=clean, norun=norun, inplace=inplace)
    model_list = model_ids if model_ids else None

    if virtual:
        from pyvirtualdisplay import Display
        with Display(manage_global_env=True, visible=False) as _:
            mrm.run_models(model_list)
    else:
        mrm.run_models(model_list)


def getmodels(args=None):
    """getmodels

    Retrieve all or specified models from ModelDB.

    Usage:
        getmodels [<model_id>...]
        getmodels -h

    Arguments:
        model_id=<n>           Optional: ModelDB accession number(s) to download; default is all available models

    Examples
        getmodels
        getmodels 23613 12344
    """
    options = docopt(getmodels.__doc__, args)
    model_ids = [int(model_id) for model_id in options.pop("<model_id>")]

    mdb = ModelDB()
    mdb.download_models(model_list=model_ids if model_ids else None)


def diffgout(args=None):
    """diffgout

        Graphically compare two gout files

    Usage:
        diffgout <goutFile1> <goutFile2>
        diffgout -h         Print help

    Arguments:
        goutFile1=PATH      Required: file path to first gout file
        goutFile2=PATH      Required: file path to second gout file

    Examples
        diffgout 3246-master/varela/gout 3246-8.0.2/varela/gout

    """
    options = docopt(diffgout.__doc__, args)

    gout_file1 = options.pop("<goutFile1>")
    gout_file2 = options.pop("<goutFile2>")

    cmd = 'nrngui -c "strdef gout1" -c "gout1=\\"{}\\"" -c "strdef gout2" -c "gout2=\\"{}\\"" modeldb/showgout.hoc'.format(
        gout_file1, gout_file2)
    commands = shlex.split(cmd)
    _ = subprocess.Popen(commands)


def modeldb_config(args=None):
    """modeldb-config

    Print out ModelDB configuration items

    Usage:
        modeldb-config [options]
        modeldb-config -h    Print help

    Options:
        --item=ARG          Print out value for specific config item

    Examples
        modeldb-config
        modeldb-config --item=MDB_NEURON_MODELS_URL
    """
    options = docopt(modeldb_config.__doc__, args)
    item = options.pop("--item", None)
    cfg_module = globals().get('config', None)
    if item is None:
        pprint({var: getattr(cfg_module, var) for var in dir(cfg_module) if
            not inspect.ismodule(var) and not var.startswith("__") and not var.endswith("__")})
    else:
        print(getattr(cfg_module, item))


def report2html(args=None):
    """report2html

        Create an interactive HTML report from a single run.

    Usage:
        report2html <json_report>
        report2html -h         Print help

    Arguments:
        json_report=PATH      Required: json report file following runmodels

    Examples
        report2html 3246-master.json

    """
    options = docopt(report2html.__doc__, args)

    json_report = options.pop("<json_report>")

    file_loader = FileSystemLoader(os.path.join(Path(__file__).parent.resolve(), 'templates'))
    env = Environment(loader=file_loader)
    template = env.get_template('report.html')

    report_filename = os.path.join(os.path.splitext(os.path.basename(json_report))[0] + '.html')
    print('Writing {} ...'.format(report_filename))
    with open(report_filename, 'w') as fh, open(json_report, 'r+') as jr:
        fh.write(template.render(
            title="{} : nr-modeldb-ci HTML report".format(json_report),
            json_report=json.load(jr),
        ))
    print('Done.')


def diffreports2html(args=None):
    """diffreports2html

        Create an interactive HTML report from two nrn-modeldb-ci json reports.
        Note that you should have the gout files present if you want to diff them.

    Usage:
        diffreports2html <json_report1> <json_report2>
        diffreports2html -h         Print help

    Arguments:
        json_report1=PATH      Required: json report file following runmodels for NEURON version 1
        json_report2=PATH      Required: json report file following runmodels for NEURON version 2

    Examples
        diffreport2html 3246-master.json 3246-8.0.2.json

    """
    options = docopt(diffreports2html.__doc__, args)

    json_report1 = options.pop("<json_report1>")
    json_report2 = options.pop("<json_report2>")

    file_loader = FileSystemLoader(os.path.join(Path(__file__).parent.resolve(), 'templates'))
    env = Environment(loader=file_loader)
    template = env.get_template('diffreport.html')
    runtime_template = env.get_template('runtimes.html')


    report_title = '{}-vs-{}'.format( os.path.splitext(os.path.basename(json_report1))[0],
                                      os.path.splitext(os.path.basename(json_report2))[0] )
    report_filename = os.path.join(Path(json_report1).resolve().parent, report_title + '.html')
    runtime_report_title = 'Runtimes ' + report_title
    runtime_report_filename = os.path.join(Path(json_report1).resolve().parent, "runtimes-" + report_title + '.html')
    diff_dict, gout_dict, runtime_dict, stats_dict, v1, v2 = diff_reports(json_report1, json_report2)

    print('Writing {} ...'.format(report_filename))
    with open(report_filename, 'w') as fh:
        fh.write(template.render(
            title="{}".format(report_title),
            diff_dict=diff_dict,
            gout_dict=gout_dict),
        )
    print('Done.')
    print('Writing {} ...'.format(runtime_report_filename))
    with open(runtime_report_filename, 'w') as fh:
        fh.write(runtime_template.render(
            title="{}".format(runtime_report_title),
            runtime_dict=runtime_dict,
            stats={"stats":diff_dict['0']},
            v1=v1,
            v2=v2),
    )
    print('Done.')
    # Return a useful status code
    code = 0
    if len(diff_dict) > 1:
        assert "0" in diff_dict  # summary info; not a real diff
        print("FAILURE: stdout diffs in {}".format(set(diff_dict.keys()) - {"0"}))
        code = 1
    if len(gout_dict) > 1:
        assert "0" in gout_dict  # summary info; not a real diff
        print("FAILURE: gout diffs in {}".format(set(diff_dict.keys()) - {"0"}))
        code = 1
    total_failures = sum(
        version_stats["Failed models"]["Count"] for version_stats in stats_dict.values()
    )
    if total_failures > 0:
        print(
            "FAILURE: there were {} failed model builds across {} versions of NEURON".format(
                total_failures, len(stats_dict)
            )
        )
        code = 1
    total_runtime_failures = sum(
        version_stats["Failed runs"]["Count"] for version_stats in stats_dict.values()
    )
    if total_runtime_failures > 0:
        print(
            "FAILURE: there were {} failed model runs across {} versions of NEURON".format(
                total_runtime_failures, len(stats_dict)
            )
        )
        code = 1
    # These are not expected to be different between the two NEURON versions tested
    assert (
        len(
            {
                version_stats["Skipped runs"]["Count"]
                for version_stats in stats_dict.values()
            }
        )
        == 1
    )
    assert (
        len(
            {
                version_stats["Total nof models run"]
                for version_stats in stats_dict.values()
            }
        )
        == 1
    )
    return code



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

install_requires = [
    # core
    'requests',
    'pyyaml',
    'docopt',
    'jinja2',
    'pyvirtualdisplay',

    # for running models
    'pyqt5',
    'ipython',
    'matplotlib',
    'scipy',
    '2to3',
]

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()


def setup_package():

    setup(
        name='nrn-modeldb-ci',
        description='NEURON ModelDB CI tools',
        url='https://github.com/neuronsimulator/nrn-modeldb-ci',
        author='EPFL Blue Brain Project & Yale',
        author_email='alexandru.savulescu@epfl.ch',
        license='BSD-3-Clause',
        packages=find_packages(),
        use_scm_version=True,
        include_package_data=True,
        install_requires=install_requires,
        setup_requires=['setuptools_scm'],
        entry_points=dict(
            console_scripts=[
                'runmodels = modeldb.commands:runmodels',
                'getmodels = modeldb.commands:getmodels',
                'diffgout = modeldb.commands:diffgout',
                'modeldb-config = modeldb.commands:modeldb_config',
                'report2html = modeldb.commands:report2html',
                'diffreports2html = modeldb.commands:diffreports2html',
            ]
        ),
        long_description=long_description,
        long_description_content_type="text/markdown",
    )


if __name__ == "__main__":
    setup_package()
